from rdflib.graph import ConjunctiveGraph as ConjunctiveGraph
from rdflib.graph import Graph as Graph
from rdflib.namespace import CSVW as CSVW
from rdflib.namespace import DC as DC
from rdflib.namespace import DCAT as DCAT
from rdflib.namespace import DCTERMS as DCTERMS
from rdflib.namespace import DOAP as DOAP
from rdflib.namespace import FOAF as FOAF
from rdflib.namespace import ODRL2 as ODRL2
from rdflib.namespace import ORG as ORG
from rdflib.namespace import OWL as OWL
from rdflib.namespace import PROF as PROF
from rdflib.namespace import PROV as PROV
from rdflib.namespace import QB as QB
from rdflib.namespace import RDF as RDF
from rdflib.namespace import RDFS as RDFS
from rdflib.namespace import SDO as SDO
from rdflib.namespace import SH as SH
from rdflib.namespace import SKOS as SKOS
from rdflib.namespace import SOSA as SOSA
from rdflib.namespace import SSN as SSN
from rdflib.namespace import TIME as TIME
from rdflib.namespace import VOID as VOID
from rdflib.namespace import XMLNS as XMLNS
from rdflib.namespace import XSD as XSD
from rdflib.namespace import Namespace as Namespace
from rdflib.term import BNode as BNode
from rdflib.term import Literal as Literal
from rdflib.term import URIRef as URIRef
from rdflib.term import Variable as Variable

__version__: str
