from rdflib.namespace import DefinedNamespace as DefinedNamespace
from rdflib.namespace import Namespace as Namespace
from rdflib.term import URIRef as URIRef

class ODRL2(DefinedNamespace):
    action: URIRef
    andSequence: URIRef
    assignee: URIRef
    assigneeOf: URIRef
    assigner: URIRef
    assignerOf: URIRef
    attributedParty: URIRef
    attributingParty: URIRef
    compensatedParty: URIRef
    compensatingParty: URIRef
    conflict: URIRef
    consentedParty: URIRef
    consentingParty: URIRef
    consequence: URIRef
    constraint: URIRef
    contractedParty: URIRef
    contractingParty: URIRef
    dataType: URIRef
    duty: URIRef
    failure: URIRef
    function: URIRef
    hasPolicy: URIRef
    implies: URIRef
    includedIn: URIRef
    informedParty: URIRef
    informingParty: URIRef
    inheritAllowed: URIRef
    inheritFrom: URIRef
    inheritRelation: URIRef
    leftOperand: URIRef
    obligation: URIRef
    operand: URIRef
    operator: URIRef
    output: URIRef
    partOf: URIRef
    payeeParty: URIRef
    permission: URIRef
    profile: URIRef
    prohibition: URIRef
    proximity: URIRef
    refinement: URIRef
    relation: URIRef
    remedy: URIRef
    rightOperand: URIRef
    rightOperandReference: URIRef
    scope: URIRef
    source: URIRef
    status: URIRef
    target: URIRef
    timedCount: URIRef
    trackedParty: URIRef
    trackingParty: URIRef
    uid: URIRef
    undefined: URIRef
    unit: URIRef
    xone: URIRef
    All: URIRef
    All2ndConnections: URIRef
    AllConnections: URIRef
    AllGroups: URIRef
    Group: URIRef
    Individual: URIRef
    absolutePosition: URIRef
    absoluteSize: URIRef
    absoluteSpatialPosition: URIRef
    absoluteTemporalPosition: URIRef
    count: URIRef
    dateTime: URIRef
    delayPeriod: URIRef
    deliveryChannel: URIRef
    device: URIRef
    elapsedTime: URIRef
    eq: URIRef
    event: URIRef
    fileFormat: URIRef
    gt: URIRef
    gteq: URIRef
    hasPart: URIRef
    ignore: URIRef
    industry: URIRef
    invalid: URIRef
    isA: URIRef
    isAllOf: URIRef
    isAnyOf: URIRef
    isNoneOf: URIRef
    isPartOf: URIRef
    language: URIRef
    lt: URIRef
    lteq: URIRef
    media: URIRef
    meteredTime: URIRef
    neq: URIRef
    payAmount: URIRef
    percentage: URIRef
    perm: URIRef
    policyUsage: URIRef
    product: URIRef
    prohibit: URIRef
    purpose: URIRef
    recipient: URIRef
    relativePosition: URIRef
    relativeSize: URIRef
    relativeSpatialPosition: URIRef
    relativeTemporalPosition: URIRef
    resolution: URIRef
    spatial: URIRef
    spatialCoordinates: URIRef
    support: URIRef
    system: URIRef
    systemDevice: URIRef
    timeInterval: URIRef
    unitOfCount: URIRef
    version: URIRef
    virtualLocation: URIRef
    Action: URIRef
    Agreement: URIRef
    Assertion: URIRef
    Asset: URIRef
    AssetCollection: URIRef
    AssetScope: URIRef
    ConflictTerm: URIRef
    Constraint: URIRef
    Duty: URIRef
    LeftOperand: URIRef
    LogicalConstraint: URIRef
    Offer: URIRef
    Operator: URIRef
    Party: URIRef
    PartyCollection: URIRef
    PartyScope: URIRef
    Permission: URIRef
    Policy: URIRef
    Privacy: URIRef
    Prohibition: URIRef
    Request: URIRef
    RightOperand: URIRef
    Rule: URIRef
    Set: URIRef
    Ticket: URIRef
    UndefinedTerm: URIRef
    acceptTracking: URIRef
    adHocShare: URIRef
    aggregate: URIRef
    annotate: URIRef
    anonymize: URIRef
    append: URIRef
    appendTo: URIRef
    archive: URIRef
    attachPolicy: URIRef
    attachSource: URIRef
    attribute: URIRef
    commercialize: URIRef
    compensate: URIRef
    concurrentUse: URIRef
    copy: URIRef
    core: URIRef
    delete: URIRef
    derive: URIRef
    digitize: URIRef
    display: URIRef
    distribute: URIRef
    ensureExclusivity: URIRef
    execute: URIRef
    export: URIRef
    extract: URIRef
    extractChar: URIRef
    extractPage: URIRef
    extractWord: URIRef
    give: URIRef
    grantUse: URIRef
    include: URIRef
    index: URIRef
    inform: URIRef
    install: URIRef
    lease: URIRef
    lend: URIRef
    license: URIRef
    modify: URIRef
    move: URIRef
    nextPolicy: URIRef
    obtainConsent: URIRef
    pay: URIRef
    play: URIRef
    present: URIRef
    preview: URIRef
    print: URIRef
    read: URIRef
    reproduce: URIRef
    reviewPolicy: URIRef
    secondaryUse: URIRef
    sell: URIRef
    share: URIRef
    shareAlike: URIRef
    stream: URIRef
    synchronize: URIRef
    textToSpeech: URIRef
    transfer: URIRef
    transform: URIRef
    translate: URIRef
    uninstall: URIRef
    use: URIRef
    watermark: URIRef
    write: URIRef
    writeTo: URIRef
