from rdflib.namespace import DefinedNamespace as DefinedNamespace
from rdflib.namespace import Namespace as Namespace
from rdflib.term import URIRef as URIRef

class DOAP(DefinedNamespace):
    audience: URIRef
    blog: URIRef
    browse: URIRef
    category: URIRef
    created: URIRef
    description: URIRef
    developer: URIRef
    documenter: URIRef
    helper: URIRef
    implements: URIRef
    language: URIRef
    license: URIRef
    location: URIRef
    maintainer: URIRef
    module: URIRef
    name: URIRef
    os: URIRef
    platform: URIRef
    release: URIRef
    repository: URIRef
    repositoryOf: URIRef
    revision: URIRef
    screenshots: URIRef
    shortdesc: URIRef
    tester: URIRef
    translator: URIRef
    vendor: URIRef
    wiki: URIRef
    ArchRepository: URIRef
    BKRepository: URIRef
    BazaarBranch: URIRef
    CVSRepository: URIRef
    DarcsRepository: URIRef
    GitBranch: URIRef
    GitRepository: URIRef
    HgRepository: URIRef
    Project: URIRef
    Repository: URIRef
    SVNRepository: URIRef
    Specification: URIRef
    Version: URIRef
    homepage: URIRef
